/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.field.shape;

import com.vividsolutions.jts.geom.Geometry;
import java.awt.Image;
import java.awt.datatransfer.Transferable;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.Action;
import javax.swing.Icon;
import javax.swing.ImageIcon;
import nl.cloudfarming.client.field.model.Shape;
import nl.cloudfarming.client.field.model.ShapeCollection;
import nl.cloudfarming.client.imp.ImportSource;
import nl.cloudfarming.client.geoviewer.BaseLayer;
import nl.cloudfarming.client.geoviewer.LayerInfo;
import nl.cloudfarming.client.geoviewer.RandomColorPalette;
import nl.cloudfarming.client.geoviewer.SimpleLayer;
import org.geotools.data.FeatureSource;
import org.geotools.data.shapefile.ShapefileDataStore;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.geotools.geometry.jts.JTS;
import org.geotools.referencing.CRS;
import org.opengis.feature.simple.SimpleFeature;
import org.opengis.feature.type.AttributeDescriptor;
import org.opengis.feature.type.AttributeType;
import org.opengis.feature.type.Name;
import org.opengis.referencing.FactoryException;
import org.opengis.referencing.NoSuchAuthorityCodeException;
import org.opengis.referencing.crs.CoordinateReferenceSystem;
import org.opengis.referencing.operation.MathTransform;
import org.opengis.referencing.operation.TransformException;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiFileLoader;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.ChildFactory;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.util.Exceptions;
import org.openide.util.ImageUtilities;
import org.openide.util.Lookup;
import org.openide.util.NbBundle;
import org.openide.util.datatransfer.PasteType;
import org.openide.util.lookup.AbstractLookup;
import org.openide.util.lookup.InstanceContent;
import org.openide.util.lookup.Lookups;
import org.openide.util.lookup.ProxyLookup;

/**
 * DataObject for ESRI shape files
 *
 */
@NbBundle.Messages("ShapeDataObject.shape_name=Unknown")
public class ShapeDataObject extends MultiDataObject {

    public static final int CONTENT_NONE = 0;
    public static final int CONTENT_FIELDS = 1;
    public static final int CONTENT_TREATMENT_ZONES = 2;
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.shapetype.ShapeDataObject");
    private static final String ATT_NAME_PROPERTY = "att_name_property";
    private static final String ATT_VALUE_PROPERTY = "att_value_property";
    private static final String ATT_CONTENT_TYPE = "att_content_type";
    private static final String ATT_SHAPE_TYPE = "att_shape_type";
    private static final String ATT_NUMBER_OF_INVALID_SHAPES = "att_number_invalid_shapes";
    private static final String ATT_NUMBER_OF_CORRECTED_SHAPES = "att_number_corrected_shapes";
    public static final String MIMETYPE = "application/x-esri-shape";
    public static final ImageIcon ICON = ImageUtilities.loadImageIcon("nl/cloudfarming/client/field/shape/file_extension_shp.png", true);
    public static final ImageIcon BADGE_ERROR = ImageUtilities.loadImageIcon("nl/cloudfarming/client/field/shape/exclamation.png", true);
    public static final ImageIcon BADGE_WARN = ImageUtilities.loadImageIcon("nl/cloudfarming/client/field/shape/exclamation2.png", true);
    public static final Icon ICON_ERROR = ImageUtilities.image2Icon(ImageUtilities.mergeImages(ICON.getImage(), BADGE_ERROR.getImage(), 0, 0));
    public static final Icon ICON_WARN = ImageUtilities.image2Icon(ImageUtilities.mergeImages(ICON.getImage(), BADGE_WARN.getImage(), 0, 0));
    private final InstanceContent ic = new InstanceContent();
    private final Lookup lkp = new AbstractLookup(ic);
    private List<Shape> shapes = null;
    private List<String> stringAttributes;
    private List<String> integerAttributes;
    private final TreatmentZoneImportCookie treatmentZoneImportCookie = new TreatmentZoneImportCookie(this);
    private final FieldImportCookie fieldImportCookie = new FieldImportCookie(this);

    public ShapeDataObject(FileObject pf, MultiFileLoader loader) throws DataObjectExistsException, IOException {
        super(pf, loader);
        registerEditor(MIMETYPE, true);
        ic.add(new ImportSource() {
            @Override
            public Set<FileObject> getFiles() {
                return files();
            }
        });
        checkDefaultAttributes();
        checkValidity();
        updateImportCookie();
    }

    /**
     * check for mandatory secondary files
     *
     * @return
     */
    private boolean isComplete() {
        if (getPrimaryFile() == null){
            return false;
        }
        for (String manditory : ShapeFileLoader.MANDITORY_EXT) {
            if (FileUtil.findBrother(getPrimaryFile(), manditory) == null) {
                return false;
            }
        }
        // all manditory extentions found
        return true;
    }

    // check shapefile attributes against registered property mappings:
    private void checkDefaultAttributes() {
        // only check for attributes when no configuration has been done yet:
        if (getContentType() == CONTENT_NONE && getNameProperty() == null && getValueProperty() == null) {
            updateMappingAttributes(ShapePropertyMappingRegistry.getInstance().findAll(), false);
        }
    }

    /**
     * Check shapefile attributes against a custom list of property mappings,
     * possibly clearing existing attributes when no match is found.
     *
     * @param mappings
     */
    public void updateMappingAttributes(List<ShapePropertyMapping> mappings, boolean reset) {

        if (reset) {
            setNameProperty(null);
            setValueProperty(null);
            setContentType(CONTENT_NONE);
            for (Shape shape : getShapes()) {
                shape.setName(null);
                shape.setValue(null);
            }
        }

        if (mappings == null || mappings.isEmpty()) {
            return;
        }

        final List<String> nameProperties = new LinkedList<>(getStringAttributes());
        nameProperties.addAll(getIntegerAttributes());
        final List<String> valueProperties = new LinkedList<>(getIntegerAttributes());
        final ShapePropertyMapping mapping = ShapePropertyMappingRegistry.findMatch(mappings, nameProperties, valueProperties);
        if (mapping != null) {
            LOGGER.log(Level.INFO, "applying shape attribute mapping: {0}", mapping);
            setValueProperty(mapping.getValueProperty());
            setContentType(CONTENT_TREATMENT_ZONES);
            for (Shape shape : getShapes()) {
                shape.setValueFromAttribute(mapping.getValueProperty());
            }
        }
    }

    private void checkValidity() {
        if (getShapeType() == null || getNumberOfInvalidShapes() == null) {
            boolean hasType = false;
            int numberOfInvalidShapes = 0;
            int numberOfCorrectedShapes = 0;

            for (Shape shape : getShapes()) {

                String type = shape.getGeometry().getGeometryType();
                // shp files contain a single geometry type, set it once:
                if (!hasType && type != null) {
                    setShapeType(type.toUpperCase());
                    hasType = true;
                }

                if (!shape.isValid()) {
                    LOGGER.warning(shape.getValidationMessage());
                    numberOfInvalidShapes++;
                } else if (shape.isAutocorrected()) {
                    numberOfCorrectedShapes++;
                }
            }
            setNumberOfInvalidShapes(numberOfInvalidShapes);
            setNumberOfCorrectedShapes(numberOfCorrectedShapes);
        }
    }

    public String getShapeType() {
        return (String) getPrimaryFile().getAttribute(ATT_SHAPE_TYPE);
    }

    private void setShapeType(String shapeType) {
        try {
            getPrimaryFile().setAttribute(ATT_SHAPE_TYPE, shapeType);
        } catch (IOException ex) {
            LOGGER.log(Level.WARNING, "could not save shape type property for shape data object {0}", getName());
        }
    }

    public Integer getNumberOfInvalidShapes() {
        return (Integer) getPrimaryFile().getAttribute(ATT_NUMBER_OF_INVALID_SHAPES);
    }

    private void setNumberOfInvalidShapes(Integer numberOfInvalidShapes) {
        try {
            getPrimaryFile().setAttribute(ATT_NUMBER_OF_INVALID_SHAPES, numberOfInvalidShapes);
        } catch (IOException ex) {
            LOGGER.log(Level.WARNING, "could not save invalid shapes property for shape data object {0}", getName());
        }
    }

    public Integer getNumberOfCorrectedShapes() {
        return (Integer) getPrimaryFile().getAttribute(ATT_NUMBER_OF_CORRECTED_SHAPES);
    }

    private void setNumberOfCorrectedShapes(Integer numberOfCorrectedShapes) {
        try {
            getPrimaryFile().setAttribute(ATT_NUMBER_OF_CORRECTED_SHAPES, numberOfCorrectedShapes);
        } catch (IOException ex) {
            LOGGER.log(Level.WARNING, "could not save corrected shapes property for shape data object {0}", getName());
        }
    }

    /**
     * Method to set the secondary files
     *
     * @param secondary the secondary files
     */
    public void setSecondaryFiles(List<FileObject> secondary) {
        for (FileObject fileObject : secondary) {
            registerEntry(fileObject);
        }
    }

    @Override
    protected Node createNodeDelegate() {
        Node delegate = super.createNodeDelegate();
        ic.add(new ShapeCollection(getShapes()));
        //FIXME move palette to object level
        ic.add(new BaseLayer(new RandomColorPalette(), getPrimaryFile().getName()));
        LayerInfo layerInfo = new LayerInfo();
        layerInfo.setShowIcons(false);
        ic.add(layerInfo);
        return new ShapeDataObjectNode(this, delegate, getPrimaryFile().getName());
    }

    void setNameProperty(String nameProperty) {
        try {
            getPrimaryFile().setAttribute(ATT_NAME_PROPERTY, nameProperty);
        } catch (IOException ex) {
            LOGGER.log(Level.WARNING, "could not save name propery mapping for shape data object {0}", getName());
        }
    }

    /**
     * Name of the attribute which value should be used for the name of the
     * shape
     *
     * @return
     */
    String getNameProperty() {
        return (String) getPrimaryFile().getAttribute(ATT_NAME_PROPERTY);
    }

    void setValueProperty(String valueProperty) {
        try {
            getPrimaryFile().setAttribute(ATT_VALUE_PROPERTY, valueProperty);
        } catch (IOException ex) {
            LOGGER.log(Level.WARNING, "could not save value propery mapping for shape data object {0}", getName());
        }
    }

    /**
     * Retrieve the value property for a shapefile. The value property is name
     * of the attribute in the shape file use to map its value to the value
     * property of the shape.
     *
     * @return
     */
    String getValueProperty() {
        return (String) getPrimaryFile().getAttribute(ATT_VALUE_PROPERTY);
    }

    void setContentType(int contentType) {
        try {
            getPrimaryFile().setAttribute(ATT_CONTENT_TYPE, contentType);
            updateImportCookie();
        } catch (IOException ex) {
            LOGGER.log(Level.WARNING, "could not save content type propery mapping for shape data object {0}", getName());
        }
    }

    private void updateImportCookie() {
        int contentType = getContentType();

        // TODO: take into account shape geometry type (should be *polygon for both content types below),
        // and shape validity, might keep both in file attributes.

        if (CONTENT_TREATMENT_ZONES == contentType && getLookup().lookup(TreatmentZoneImportCookie.class) == null) {
            ic.add(treatmentZoneImportCookie);
        } else if (CONTENT_TREATMENT_ZONES != contentType && getLookup().lookup(TreatmentZoneImportCookie.class) != null) {
            ic.remove(treatmentZoneImportCookie);
        }

        if (CONTENT_FIELDS == contentType && getLookup().lookup(FieldImportCookie.class) == null) {
            ic.add(fieldImportCookie);
        } else if (CONTENT_FIELDS != contentType && getLookup().lookup(FieldImportCookie.class) != null) {
            ic.remove(fieldImportCookie);
        }
    }

    /**
     * Get the content type for this shapefile <p>possible values: <ul>
     * <li>{@link ShapeDataObject#CONTENT_NONE}</li>
     * <li>{@link ShapeDataObject#CONTENT_FIELDS}</li>
     * <li>{@link ShapeDataObject#CONTENT_TREATMENT_ZONES}</li> <ul>
     *
     * </p>
     *
     * @return
     */
    public int getContentType() {
        Integer contentType = (Integer) getPrimaryFile().getAttribute(ATT_CONTENT_TYPE);
        return contentType == null ? CONTENT_NONE : contentType.intValue();
    }

    @Override
    public Lookup getLookup() {
        return new ProxyLookup(super.getLookup(), lkp);
    }

    /**
     * Retrieve all shapes in this shapeDataObject
     *
     * @return
     */
    public final List<Shape> getShapes() {
        if (shapes == null) {
            shapes = new ArrayList<>();
            try {
                ShapefileDataStore shpDataStore = new ShapefileDataStore(getPrimaryFile().toURL());
                String typeName = shpDataStore.getTypeNames()[0];
                FeatureSource featureSource = shpDataStore.getFeatureSource(typeName);

                CoordinateReferenceSystem wgs84 = CRS.decode("EPSG:4326", true);
                CoordinateReferenceSystem shapeCRS = shpDataStore.getSchema().getCoordinateReferenceSystem();
                MathTransform transform = null;
                if (shapeCRS == null) {
                    LOGGER.warning("Shape did not contain a CRS, assuming EPSG:4326 (WGS84), mileage may vary");
                } else {
                    boolean lenient = true; // allow for some error due to different datums
                    transform = CRS.findMathTransform(shapeCRS, wgs84, lenient);
                    LOGGER.log(Level.FINEST, "created transform from {0} to {1}", new Object[]{shapeCRS, wgs84});
                }

                FeatureCollection result = featureSource.getFeatures();
                FeatureIterator iterator = result.features();
                while (iterator.hasNext()) {
                    SimpleFeature feature = (SimpleFeature) iterator.next();
                    Shape shape = getShape(feature, transform);
                    //
                    // parse all available string attributes
                    //
                    stringAttributes = readAttributes(feature, String.class);
                    //
                    // add all string attributes and their values to the map in the shape object
                    //
                    for (String att : stringAttributes) {
                        int index = feature.getFeatureType().indexOf(att);
                        if (index >= 0) {
                            String value = (String) feature.getAttribute(index);
                            shape.addStringAttribute(att, value);
                        }
                    }
                    //
                    // parse all available Integer attributes
                    //
                    integerAttributes = readAttributes(feature, Integer.class);
                    //
                    // add all string attributes and their values to the map in the shape object
                    //
                    for (String att : integerAttributes) {
                        int index = feature.getFeatureType().indexOf(att);
                        if (index >= 0) {
                            Integer value = (Integer) feature.getAttribute(index);
                            shape.addIntegerAttribute(att, value == null ? null : value.doubleValue());
                        }
                    }

                    List<String> longValues = readAttributes(feature, Long.class);
                    for (String att : longValues) {
                        int index = feature.getFeatureType().indexOf(att);
                        if (index >= 0) {
                            Long value = (Long) feature.getAttribute(index);
                            shape.addIntegerAttribute(att, value == null ? null : value.doubleValue());
                        }
                    }
                    integerAttributes.addAll(longValues);

                    List<String> doubleValues = readAttributes(feature, Double.class);
                    for (String att : doubleValues) {
                        int index = feature.getFeatureType().indexOf(att);
                        if (index >= 0) {
                            Double value = (Double) feature.getAttribute(index);
                            shape.addIntegerAttribute(att, value);
                        }
                    }
                    integerAttributes.addAll(doubleValues);
                    shape.setNameFromAttribute(getNameProperty());
                    shape.setValueFromAttribute(getValueProperty());

                    shapes.add(shape);
                }
                iterator.close();
            } catch (TransformException | FactoryException | IOException ex) {
                LOGGER.log(Level.WARNING, "Could not return shapes", ex.getMessage());
                Exceptions.printStackTrace(ex);
            }
        }
        return shapes;
    }

    /**
     * Retrieve all valid shapes in this shapeDataObject.
     *
     * @return
     */
    public List<Shape> getValidShapes() {
        List<Shape> validShapes = new ArrayList<>();
        for (Shape s : getShapes()) {
            if (s.isValid()) {
                validShapes.add(s);
            }
        }
        return validShapes;
    }

    /**
     * Retrieve all invalid shapes in this shapeDataObject.
     *
     * @return
     */
    public List<Shape> getInvalidShapes() {
        List<Shape> invalidShapes = new ArrayList<>();
        for (Shape s : getShapes()) {
            if (!s.isValid()) {
                invalidShapes.add(s);
            }
        }
        return invalidShapes;
    }

    /**
     * Retrieve all shapes in this shapeDataObject that had an autocorrection
     * applied.
     *
     * @return
     */
    public List<Shape> getAutocorrectedShapes() {
        List<Shape> autocorrectedShapes = new ArrayList<>();
        for (Shape s : getShapes()) {
            if (s.isAutocorrected()) {
                autocorrectedShapes.add(s);
            }
        }
        return autocorrectedShapes;
    }

    /**
     * Reads attributes of type String from specified feature into the
     * attributes List
     *
     * @param feature
     */
    private List<String> readAttributes(SimpleFeature feature, Class clazz) {
        List<String> returnList = new ArrayList<>();
        List<AttributeDescriptor> attributeDescriptors = feature.getFeatureType().getAttributeDescriptors();
        for (AttributeDescriptor attributeDescriptor : attributeDescriptors) {
            // Determine type
            AttributeType attributeType = attributeDescriptor.getType();

            boolean attributeIsString = attributeType.getBinding().equals(clazz);

            if (attributeIsString) {
                // Determine Attributename
                Name name = attributeDescriptor.getName();
                if (name != null) {
                    returnList.add(name.toString());
                }
            }
        }
        return returnList;
    }

    /**
     * Get the attributes of type 'Integer' from this shp-file
     *
     * @return The attributes inside the shp-file
     */
    public List<String> getIntegerAttributes() {
        if (integerAttributes == null) {
            getShapes();
        }
        return integerAttributes;
    }

    /**
     * Get the attributes of type 'String' from this shp-file
     *
     * @return The attributes inside the shp-file
     */
    public List<String> getStringAttributes() {
        // Read shapes if not already done
        if (stringAttributes == null) {
            getShapes();
        }
        return stringAttributes;
    }

    /**
     *
     * @param payload
     * @return
     */
    private Shape getShape(SimpleFeature payload, MathTransform transform) throws TransformException, NoSuchAuthorityCodeException, FactoryException {

        int attIndexPlotPoly = payload.getFeatureType().indexOf("the_geom");

        Geometry geometry = null;
        if (attIndexPlotPoly >= 0) {
            geometry = (Geometry) payload.getAttribute(attIndexPlotPoly);
            if (transform != null) {
                LOGGER.log(Level.FINEST, "transforming geometry {0}", geometry);
                geometry = JTS.transform(geometry, transform);
                LOGGER.log(Level.FINEST, "transformed geometry {0}", geometry);
            }
        }
        String shapeName = Bundle.ShapeDataObject_shape_name();

        Shape shape = new Shape(geometry);
        shape.setName(shapeName);
        return shape;
    }

    @Override
    protected int associateLookup() {
        return 1;
    }

    /**
     * Root node for the shapefile
     */
    @NbBundle.Messages({
        "# {0} - shape file name",
        "ShapeDataObjectNode invalid shape file display name={0} (invalid)",})
    private final class ShapeDataObjectNode extends AbstractNode {

        final Node original;
        private final ShapeDataObject shapeDataObject;
        private final String name;
        private final Image ICON = ImageUtilities.loadImage("nl/cloudfarming/client/field/shape/file_extension_shp.png");

        public ShapeDataObjectNode(ShapeDataObject shapeDataObject, Node original, String name) {
            super(Children.create(new ShapeChildFactory(), true), original.getLookup());
            this.original = original;
            this.shapeDataObject = shapeDataObject;
            this.name = name;
        }

        @Override
        public Action[] getActions(boolean context) {
            return original.getActions(context);
        }

        @Override
        public PasteType getDropType(Transferable t, int action, int index) {
            return null;
        }

        @Override
        public boolean canDestroy() {
            return original.canDestroy();
        }

        @Override
        public void destroy() throws IOException {
            original.destroy();
        }

        @Override
        public String getDisplayName() {
            return shapeDataObject.getNumberOfInvalidShapes() > 0 ? Bundle.ShapeDataObjectNode_invalid_shape_file_display_name(name) : name;
        }

        @Override
        public Image getIcon(int type) {
            return getDefaultIcon();
        }

        @Override
        public Image getOpenedIcon(int type) {
            return getDefaultIcon();
        }

        private Image getDefaultIcon() {
            if (shapeDataObject.getNumberOfInvalidShapes() > 0) {
                return ImageUtilities.icon2Image(ICON_ERROR);
            } else if (shapeDataObject.getNumberOfCorrectedShapes() > 0) {
                return ImageUtilities.icon2Image(ICON_WARN);
            }
            return ICON;
        }
    }

    /**
     * Creates shape nodes for the simple features within the specified
     * shapefile
     */
    private final class ShapeChildFactory extends ChildFactory<Shape> {

        @Override
        protected boolean createKeys(List<Shape> toPopulate) {
            // Add all features in this Dataobject to the nodelist
            toPopulate.addAll(ShapeDataObject.this.getShapes());
            return true;
        }

        @Override
        protected Node createNodeForKey(Shape key) {
            Node shapeNode = new ShapeNode(key);
            return shapeNode;
        }
    }

    /**
     * Node representing a SimpleFeature (shape) in a shape file
     */
    private final class ShapeNode extends AbstractNode {

        private final Shape shape;

        public ShapeNode(Shape shape) {
            // AGROSENSE-1163: add the containing data object to the lookup to link the shape to a project
            super(Children.LEAF, Lookups.fixed(new ShapeCollection(shape), shape,
                    new RandomColorPalette(shape.isValid(), shape.isAutocorrected()), new SimpleLayer(shape.getName()), ShapeDataObject.this));
            this.shape = shape;
            setName(shape.getName());
            shape.addPropertyChangeListener(new PropertyChangeListener() {
                @Override
                public void propertyChange(PropertyChangeEvent evt) {
                    if (Shape.PROP_NAME.equals(evt.getPropertyName())) {
                        setName((String) evt.getNewValue());
                    }
                }
            });

        }

        @Override
        public PasteType getDropType(Transferable t, int action, int index) {
            return null;
        }

        @Override
        public Image getIcon(int type) {
            return shape.getIcon();
        }

        @Override
        public Image getOpenedIcon(int type) {
            return shape.getIcon();
        }
    }
}
