/**
 * Copyright (C) 2008-2012 AgroSense Foundation.
 *
 * AgroSense is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * There are special exceptions to the terms and conditions of the GPLv3 as it
 * is applied to this software, see the FLOSS License Exception
 * <http://www.agrosense.eu/foss-exception.html>.
 *
 * AgroSense is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
 * A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * AgroSense. If not, see <http://www.gnu.org/licenses/>.
 */
package nl.cloudfarming.client.field.shape;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.openide.awt.ActionID;
import org.openide.awt.ActionReference;
import org.openide.awt.ActionReferences;
import org.openide.awt.NotificationDisplayer;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileUtil;
import org.openide.filesystems.MIMEResolver;
import org.openide.loaders.DataObject;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.FileEntry;
import org.openide.loaders.MultiDataObject;
import org.openide.loaders.MultiDataObject.Entry;
import org.openide.loaders.MultiFileLoader;
import org.openide.util.NbBundle;

/**
 *
 * @author Timon Veenstra
 */
@ActionReferences(value = {
    @ActionReference(id =
            @ActionID(category = "System", id = "org.openide.actions.OpenAction"), path = "Loaders/"+ShapeDataObject.MIMETYPE+"/Actions", position = 100, separatorAfter = 400),
    @ActionReference(id =
            @ActionID(category = "Edit", id = "org.openide.actions.DeleteAction"), path = "Loaders/"+ShapeDataObject.MIMETYPE+"/Actions", position = 500),
    @ActionReference(id =
            @ActionID(category = "System", id = "org.openide.actions.RenameAction"), path = "Loaders/"+ShapeDataObject.MIMETYPE+"/Actions", position = 600, separatorAfter = 700),
    @ActionReference(id =
            @ActionID(category = "System", id = "org.openide.actions.ToolsAction"), path = "Loaders/"+ShapeDataObject.MIMETYPE+"/Actions", position = 1200)})
@DataObject.Registration(displayName = "#shape type name", mimeType = ShapeDataObject.MIMETYPE)
@MIMEResolver.Registration(displayName = "#shape type name", resource = "ShapeResolver.xml")
@NbBundle.Messages({
    "shape type name=ESRI Shape",
    "shape missing required file title=Shape file is missing required file",
    "# {0} - file name",
    "# {1} - missing file extention",
    "shape missing required file message=The imported shape file {0} is missing a required file (.{1}), the file will be skipped"
})
public class ShapeFileLoader extends MultiFileLoader {
    /*
     * Manditory files : 
     * .shp — shape format; the feature geometry itself 
     * .shx —shape index format; a positional index of the feature geometry to allow
     * seeking forwards and backwards quickly 
     * .dbf — attribute format; columnar attributes for each shape, in dBase IV format 
     * 
     * Optional files : 
     * .prj — projection format; the coordinate system and projection information, a
     * plain text file describing the projection using well-known text format
     * .sbn and .sbx — a spatial index of the features 
     * .fbn and .fbx — a spatial index of the features for shapefiles that are read-only 
     * .ain and .aih — an attribute index of the active fields in a table or a theme's attribute
     * table 
     * .ixs — a geocoding index for read-write shapefiles 
     * .mxs — a geocoding index for read-write shapefiles (ODB format) 
     * .atx — an attribute index for the .dbf file in the form of shapefile.columnname.atx
     * (ArcGIS 8 and later) 
     * .shp.xml — geospatial metadata in XML format, such as ISO 19115 or other schemas 
     * .cpg — used to specify the code page (only for .dbf) for identifying the character 
     * encoding to be used
     */

    public static final String PRIMARY_EXT = "shp";
    public static final String[] MANDITORY_EXT = {"shx", "dbf"};
    //TODO shp.xml is composite extention and won't be recognized
    public static final String[] OPTIONAL_EXT = {"prj", "sbn", "sbx", "fbn", "fbx", "ain", "aih", "ixs", "mxs", "atx", "shp.xml", "cpg"};
    private static final Logger LOGGER = Logger.getLogger("nl.cloudfarming.client.field");

    static {
        Arrays.sort(MANDITORY_EXT);
        Arrays.sort(OPTIONAL_EXT);
    }

    public ShapeFileLoader() {
        super(ShapeDataObject.class.getName());
    }

    @Override
    protected String actionsContext() {
        return "Loaders/" + ShapeDataObject.MIMETYPE + "/Actions/"; // NOI18N
    }

    @Override
    protected FileObject findPrimaryFile(FileObject fo) {
        // never recognize folders.
        if (fo.isFolder()) {
            return null;
        }
        String ext = fo.getExt();

        // check if (secondary) file has one of the secondary extentions
        if (Arrays.binarySearch(MANDITORY_EXT, ext) >= 0 || Arrays.binarySearch(OPTIONAL_EXT, ext) >= 0 || Arrays.binarySearch(OPTIONAL_EXT, getDoubleExtension(fo)) >= 0) {
            LOGGER.log(Level.FINEST, "FileObject {0} has secondary extention, creating secondary entry and returing find brother with primary", fo);

            return FileUtil.findBrother(fo, PRIMARY_EXT);
        }

        FileObject shapeFile = findShapePrimaryFile(fo);
        // perform check if manditory brothers exist
        if (shapeFile != null) {
            for (String manditory : MANDITORY_EXT) {
                if (FileUtil.findBrother(shapeFile, manditory) == null) {
                    LOGGER.log(Level.WARNING, "Failed to find brother manditory extention {0} for FileObject {1}", new Object[]{manditory, fo});
                    NotificationDisplayer.getDefault().notify(Bundle.shape_missing_required_file_title(),ShapeDataObject.ICON_WARN, Bundle.shape_missing_required_file_message(FileUtil.getFileDisplayName(shapeFile),manditory), null);
                    return null;
                }
            }
            // all manditory extentions found
            return shapeFile;
        }
        LOGGER.log(Level.FINEST, "No primary file found for FileObject {0}", fo);
        return null;
    }

    @Override
    protected MultiDataObject createMultiObject(FileObject primaryFile) throws DataObjectExistsException, IOException {
        List<FileObject> secondary = new ArrayList<>();
        for (String manditory : MANDITORY_EXT) {
            FileObject found = FileUtil.findBrother(primaryFile, manditory);
            if (found != null) {
                secondary.add(found);
            }
        }
        for (String optional : OPTIONAL_EXT) {
            FileObject found = FileUtil.findBrother(primaryFile, optional);
            if (found != null) {
                secondary.add(found);
            }
        }
        ShapeDataObject dataObject = new ShapeDataObject(primaryFile, this);
        dataObject.setSecondaryFiles(secondary);
        return dataObject;
    }

    @Override
    protected Entry createPrimaryEntry(MultiDataObject obj, FileObject primaryFile) {
        Entry entry = new FileEntry(obj, primaryFile);
        LOGGER.log(Level.FINEST, "Creating primary entry for {0}", primaryFile.getName());
        return entry;
    }

    @Override
    protected Entry createSecondaryEntry(MultiDataObject obj, FileObject secondaryFile) {
        assert Arrays.binarySearch(MANDITORY_EXT, secondaryFile.getExt()) >= 0 || Arrays.binarySearch(OPTIONAL_EXT, secondaryFile.getExt()) >= 0 || Arrays.binarySearch(OPTIONAL_EXT, getDoubleExtension(secondaryFile)) >= 0;
        FileEntry entry = new FileEntry(obj, secondaryFile);
        LOGGER.log(Level.FINEST, "Creating secondary entry for {0}", secondaryFile.getName());
        return entry;
    }

    private FileObject findShapePrimaryFile(FileObject fo) {
        if (fo.getExt().equals(PRIMARY_EXT)) {
            return fo;
        }
        return null;
    }

    /**
     * used to get double extensions from a fileObject. Sometimes there are more
     * dot's in a filename this method is able to return the ext of these type
     * of files
     *
     * @param fileObject the fileOject to check the extension from
     * @return The extension if found or a an empty string if not
     */
    protected String getDoubleExtension(FileObject fileObject) {
        //test if there are more dot's "." in a fileName i.e .shp.xml
        if (fileObject.getNameExt().split("\\.").length > 2) {
            return fileObject.getNameExt().substring(fileObject.getNameExt().indexOf(".") + 1);
        }
        return "";
    }
}
